unit mckMapMem;

interface

uses
	Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
	StdCtrls, Dialogs, mirror, KOL;

type
	TOpenMapWhen = (omManual, omOnCreate, omOnShow, omOnActivate, omOnPaint);

        PKOLMapMem =^TKOLMapMem;
	TKOLMapMem = class(TKOLObj)
	private
		FMapHandle, FMutexHandle, FAppListHandle : THandle;
		FMapName, FSynchMessage, FApplistMessage : string;
		FMapStrings, FAppListStrings : TStringList;
		FSize, FMessageID, FApplistMessageID : DWord;
		FMapPointer, FAppListPointer : PChar;
		FOpenMapWhen : TOpenMapWhen;
		FLocked, FIsMapOpen, FExistsAlready : Boolean;
		FReading, FAppListReading, FAutoSynch, HasOpened : Boolean;
		FOnUpdate, FOnAppListChange, FOnOpenMap, FOnCloseMap : TOnEvent;
		FOnOpenFirst, FOnCloseLast : TOnEvent;
		FFormHandle: Hwnd;
		FPNewWndHandler, FPOldWndHandler: Pointer;
		FMapsOpen : integer;
		function OpenMap0(Name0, Message0 : string; var Handle0 : THandle;
			var Pointer0 : PChar; Size0 : DWord; var MessageID0 : DWord) : Boolean;
		function GetValues(Name : string) : string;
		procedure SetValues(Name : string; const Value : string);
		procedure CloseMap0(var Handle0 : THandle;var Pointer0 : PChar);
		procedure WriteMap0(Pointer0 : PChar; Strings0 : TStringList; MessageID0, Size0 : DWord);
		procedure WriteAppListMap;
		procedure ReadAppListMap;
		procedure SetMapName(Value : String);
		procedure SetMapStrings(Value : TStringList);
		procedure SetSize(Value : DWord);
		procedure SetAutoSynch(Value : Boolean);
		procedure EnterCriticalSection;
		procedure LeaveCriticalSection;
		procedure MapStringsChange(Sender : TObject);
		procedure AppListStringsChange(Sender : TObject);
		procedure NewWndProc(var FMessage : TMessage);
		procedure Dummy(Value : string);
	protected
                procedure SetupFirst( SL: TStringList; const AName, AParent, Prefix: String ); override;
                procedure SetupLast( SL: TStringList; const AName, AParent, Prefix: String ); override;
                procedure AssignEvents( SL: TStringList; const AName: String ); override;
	public
		constructor Create(AOwner: TComponent); override;
		destructor Destroy; override;
		procedure Loaded; override;
		procedure OpenMap;
		procedure CloseMap;
		procedure ReadMap;
		procedure WriteMap;
		property ExistsAlready : Boolean read FExistsAlready;
		property IsMapOpen : Boolean read FIsMapOpen;
		property AppList : TStringList read FAppListStrings;
		property MapsOpen : integer read FMapsOpen;
		property Values[Name : string] : string read GetValues write SetValues; default;
	published
		property MaxSize : DWord read FSize write SetSize;
		property AutoSynchronize : Boolean read FAutoSynch write SetAutoSynch default True;
		property MapName : string read FMapName write SetMapName;
		property MapStrings : TStringList read FMapStrings write SetMapStrings;
		property OnUpdate : TOnEvent read FOnUpdate write FOnUpdate;
		property OnAppListChange : TOnEvent read FOnAppListChange write FOnAppListChange;
		property OnOpenMap : TOnEvent read FOnOpenMap write FOnOpenMap;
		property OnCloseMap : TOnEvent read FOnCloseMap write FOnCloseMap;
		property OnOpenFirst : TOnEvent read FOnOpenFirst write FOnOpenFirst;
		property OnCloseLast : TOnEvent read FOnCloseLast write FOnCloseLast;
		property OpenMapWhen : TOpenMapWhen read FOpenMapWhen write FOpenMapWhen default omOnShow;
	end;

const
	FAppListSize = 2000;

procedure Register;

implementation

constructor TKOLMapMem.Create(AOwner: TComponent);
var
	t : integer;
	TempMapName : string;
begin
	inherited Create(AOwner);
	FAutoSynch := True;
	FOpenMapWhen := omOnShow;
	HasOpened := False;
	FSize := 4096;
	FLocked := False;
	FIsMapOpen := False;
	FExistsAlready := False;
	FReading := False;
	FAppListReading := False;
	FMapStrings := TStringList.Create;
	FMapStrings.OnChange := MapStringsChange;
	FAppListStrings := TStringList.Create;
	with FAppListStrings do
	begin
		Sorted := True;
		Duplicates := dupIgnore;
		OnChange := AppListStringsChange;
	end;
	TempMapName := 'KOL-Map-Mem-';
	for t := 0 to 10 do TempMapName := TempMapName + Chr(Random(27) + 65);
	SetMapName(TempMapName);
	if AOwner is TForm then
	begin
		FFormHandle := (AOwner as TForm).Handle;
		FPOldWndHandler := Ptr(GetWindowLong(FFormHandle, GWL_WNDPROC));
		FPNewWndHandler :=	MakeObjectInstance(NewWndProc);
		if FPNewWndHandler = nil then Raise Exception.Create('Out of resources');
		SetWindowLong(FFormHandle, GWL_WNDPROC, Longint(FPNewWndHandler));
	end
	else Raise Exception.Create('Owner must be TForm');
end;

procedure TKOLMapMem.Loaded;
begin
	inherited Loaded;
	if (not (csDesigning in ComponentState)) and (FOpenMapWhen = omOnCreate) then OpenMap;
end;

destructor TKOLMapMem.Destroy;
begin
	CloseMap;
	SetWindowLong(FFormHandle, GWL_WNDPROC, Longint(FPOldWndHandler));
	if FPNewWndHandler <> nil then FreeObjectInstance(FPNewWndHandler);
	FAppListStrings.Free;
	FAppListStrings := nil;
	FMapStrings.Free;
	FMapStrings := nil;
	inherited destroy;
end;

procedure CreatelpSA(var SD: SECURITY_DESCRIPTOR; var SA: _SECURITY_ATTRIBUTES);
begin
   InitializeSecurityDescriptor(
      @SD,
      SECURITY_DESCRIPTOR_REVISION);
   SetSecurityDescriptorDacl(
      @SD,
      TRUE,
      0,
      FALSE);
   SA.nLength := SizeOf(_SECURITY_ATTRIBUTES);
   SA.lpSecurityDescriptor := @SD;
   SA.bInheritHandle := False;
end;

function TKOLMapMem.OpenMap0(Name0, Message0 : string; var Handle0 : THandle;
var Pointer0 : PChar; Size0 : DWord;
var MessageID0 : DWord) : Boolean;
var TempMessage : array[0..255] of Char;
    lpSD:  SECURITY_DESCRIPTOR;
    lpSA: _SECURITY_ATTRIBUTES;

begin
        CreatelpSA(lpSD, lpSA);
	Handle0 := CreateFileMapping($FFFFFFFF, @lpSA, PAGE_READWRITE or SEC_COMMIT or SEC_NOCACHE,
		0, Size0, PChar(Name0));
	if (Handle0 = INVALID_HANDLE_VALUE) or (Handle0 = 0)
		then Raise Exception.Create('Unable to create file mapping!')
	else
	begin
		if (Handle0 <> 0) and (GetLastError = ERROR_ALREADY_EXISTS)
			then FExistsAlready := True;
		Pointer0 := MapViewOfFile(Handle0, FILE_MAP_ALL_ACCESS, 0, 0, 0);
		if Pointer0 = nil then Raise Exception.Create('Unable to map view of buffer!')
		else
		begin
			StrPCopy(TempMessage, Message0);
			MessageID0 := RegisterWindowMessage(TempMessage);
			if MessageID0 = 0 then Raise Exception.Create('Could not create message!')
			else Result := True;
		end
	end;
end;

procedure TKOLMapMem.OpenMap;
var
    lpSD:  SECURITY_DESCRIPTOR;
    lpSA: _SECURITY_ATTRIBUTES;
begin
	if (FMapHandle = 0) and (FMapPointer = nil) then
	begin
		FExistsAlready := False;
		if OpenMap0(FMapName, FSynchMessage, FMapHandle, FMapPointer, FSize, FMessageID) then
		begin
                        CreatelpSA(lpSD, lpSA);
			FMutexHandle := Windows.CreateMutex(@lpSA, False, PChar(FMapName + '.Mtx'));
			if FMutexHandle = 0 then Raise Exception.Create('Unable to create Mutex!');
			if OpenMap0(FMapName + '-AppList', FApplistMessage, FAppListHandle,
				FAppListPointer, FAppListSize, FApplistMessageID) then
			begin
				FIsMapOpen := True;
				HasOpened := True;
				if FExistsAlready then ReadAppListMap
				else if Assigned(FOnOpenFirst) then FOnOpenFirst(@Self);
				FAppListStrings.Add(IntToStr(FFormHandle));
				if FExistsAlready then ReadMap
				else WriteMap;
				if Assigned(FOnOpenMap) then FOnOpenMap(@Self);
			end;
		end;
	end;
end;

procedure TKOLMapMem.CloseMap0(var Handle0 : THandle;var Pointer0 : PChar);
begin
	if Pointer0 <> nil then
	begin
		UnMapViewOfFile(Pointer0);
		Pointer0 := nil;
	end;
	if Handle0 <> 0 then
	begin
		CloseHandle(Handle0);
		Handle0 := 0;
	end;
end;

procedure TKOLMapMem.CloseMap;
begin
	if FIsMapOpen then
	begin
		if FMutexHandle <> 0 then
		begin
			FAppListStrings.Delete(FAppListStrings.IndexOf(IntToStr(FFormHandle)));
			if not (csDestroying in ComponentState) and (FApplistStrings.Count = 0)
				and Assigned(FOnCloseLast) then FOnCloseLast(@Self);
			CloseHandle(FMutexHandle);
			FMutexHandle := 0;
		end;
		CloseMap0(FAppListHandle, FAppListPointer);
		CloseMap0(FMapHandle, FMapPointer);
		FIsMapOpen := False;
		if not (csDestroying in ComponentState) and Assigned(FOnCloseMap)
			then FOnCloseMap(@Self);
	end;
end;

procedure TKOLMapMem.SetMapName(Value : string);
begin
   if (FMapName <> Value) and (FMapHandle = 0) and (Length(Value) < 246) then
   begin
	FMapName := Value;
	FSynchMessage := FMapName + '-Synch-Now';
	FAppListMessage := FMapName + '-Handles';
        Change;
   end;
end;

procedure TKOLMapMem.SetMapStrings(Value : TStringList);
begin
   if Value.Text <> FMapStrings.Text then
   begin
	if Length(Value.Text) <= FSize then FMapStrings.Assign(Value)
	else Raise Exception.Create('Can''t write Strings. Strings are too large!');
        Change;
   end;
end;

procedure TKOLMapMem.MapStringsChange(Sender : TObject);
begin
	if not (csDestroying in ComponentState) then
	begin
		if FReading and Assigned(FOnUpdate) then FOnUpdate(@Self)
		else if (not FReading) and FIsMapOpen and FAutoSynch then WriteMap;
	end;
end;

procedure TKOLMapMem.AppListStringsChange(Sender : TObject);
begin
	FMapsOpen := FAppListStrings.Count;
	if (not FAppListReading) and FIsMapOpen then WriteAppListMap;
	if (not (csDestroying in ComponentState)) and Assigned(FOnAppListChange)
		then FOnAppListChange(@Self);
end;

procedure TKOLMapMem.SetSize(Value : DWord);
var
	StringsPointer : PChar;
begin
   if (FSize <> Value) and (FMapHandle = 0) then
   begin
	StringsPointer := FMapStrings.GetText;
	if (Value < StrLen(StringsPointer) + 1) then FSize := StrLen(StringsPointer) + 1
	else FSize := Value;
	if FSize < 32 then FSize := 32;
	StrDispose(StringsPointer);
        Change;
   end;
end;

procedure TKOLMapMem.SetAutoSynch(Value : Boolean);
begin
   if FAutoSynch <> Value then
   begin
	FAutoSynch := Value;
	if FAutoSynch and FIsMapOpen then WriteMap;
        Change;
   end;
end;

procedure TKOLMapMem.ReadMap;
begin
	FReading := True;
	if (FMapPointer <> nil) then FMapStrings.SetText(FMapPointer);
	FReading := False;
end;

procedure TKOLMapMem.ReadAppListMap;
begin
	FAppListReading := True;
	if (FApplistPointer <> nil) then FAppListStrings.SetText(FApplistPointer);
	FAppListReading := False;
end;

procedure TKOLMapMem.WriteMap0(Pointer0 : PChar; Strings0 : TStringList; MessageID0, Size0 : DWord);
var
	StringsPointer : PChar;
	HandleCounter : integer;
	SendToHandle : HWnd;
begin
	if Pointer0 <> nil then
	begin
		StringsPointer := Strings0.GetText;
		EnterCriticalSection;
		if StrLen(StringsPointer) + 1 <= Size0
			then System.Move(StringsPointer^, Pointer0^, StrLen(StringsPointer) + 1)
		else Raise Exception.Create('Can''t write Strings. Strings are too large!');
		LeaveCriticalSection;
		StrDispose(StringsPointer);
		for HandleCounter := 0 to FAppListStrings.Count - 1 do
		begin
			SendToHandle := StrToInt(FAppListStrings[HandleCounter]);
			if SendToHandle <> FFormHandle then PostMessage(SendToHandle,
				MessageID0,	FFormHandle, 0);
		end;
	end;
end;

procedure TKOLMapMem.WriteMap;
begin
	WriteMap0(FMapPointer, FMapStrings, FMessageID, FSize);
end;

procedure TKOLMapMem.WriteAppListMap;
begin
	WriteMap0(FAppListPointer, FAppListStrings, FAppListMessageID, FAppListSize);
end;

procedure TKOLMapMem.EnterCriticalSection;
begin
	if (FMutexHandle <> 0) and not FLocked then
	begin
		FLocked := (WaitForSingleObject(FMutexHandle, INFINITE) = WAIT_OBJECT_0);
	end;
end;

procedure TKOLMapMem.LeaveCriticalSection;
begin
	if (FMutexHandle <> 0) and FLocked then
	begin
		ReleaseMutex(FMutexHandle);
		FLocked := False;
	end;
end;

function TKOLMapMem.GetValues(Name : string) : string;
begin
	Result := FMapStrings.Values[Name];
end;

procedure TKOLMapMem.SetValues(Name : string; const Value : string);
begin
	if Value <> FMapStrings.Values[Name] then FMapStrings.Values[Name] := Value;
end;

procedure	TKOLMapMem.NewWndProc(var FMessage : TMessage);
begin
	with FMessage do
	begin
		if not (csDesigning in ComponentState) then
		begin
		case Msg of
			WM_SHOWWINDOW: if (WParam <> 0) and (FOpenMapWhen = omOnShow)
				and (not HasOpened) then OpenMap;
			WM_ACTIVATE: if (WParam <> WA_INACTIVE) and (FOpenMapWhen = omOnActivate)
				and (not HasOpened) then OpenMap;
			WM_PAINT: if (FOpenMapWhen = omOnPaint) and (not HasOpened) then OpenMap;
		end;
		end;
		if FIsMapOpen then
		begin
			if Msg = FMessageID then ReadMap
			else if Msg = FAppListMessageID then ReadAppListMap;
		end;
		Result := CallWindowProc(FPOldWndHandler,	FFormHandle, Msg, wParam, lParam);
	end;
end;

procedure TKOLMapMem.Dummy(Value : string);
begin
		//read only
end;

procedure TKOLMapMem.SetupFirst(SL: TStringList; const AName,
  AParent, Prefix: String);
var RName: string;
        i: integer;
begin
  SL.Add( Prefix + 'if not Result.Form.HandleAllocated');
  SL.Add( Prefix + 'then Result.Form.CreateWindow;');
  SL.Add( Prefix + AName + ' := NewMapMem( Result.Form );' );
  SL.Add( Prefix + AName + '.MapName     := ''' + MapName + ''';');
  case OpenMapWhen of
  omOnActivate: RName := 'omOnActivate';
  omOnCreate: RName := 'omOnCreate';
  omOnPaint: RName := 'omOnPaint';
  omOnShow: RName := 'omOnShow';
  omManual: RName := 'omManual';
  end;
  SL.Add( Prefix + AName + '.OpenMapWhen := ' + RName + ';');
  if not FAutoSynch then
  SL.Add( Prefix + AName + '.AutoSynch   := False;');
  if FSize <> 4096 then
  SL.Add( Prefix + AName + '.Size        := ' + IntToStr(FSize) + ';');
  for i := 1 to MapStrings.Count do begin
     SL.Add( Prefix + AName + '.MapStrings.Add(''' + MapStrings[i - 1] + ''');');
  end;
  SL.Add( Prefix + AName + '.OpenMap;');
  Sl.Add( Prefix + 'if ' + AName + '.ExistsAlready then ' + AName + '.ReadMap');
  Sl.Add( Prefix + 'else ' + AName + '.WriteMap;');;
end;

procedure TKOLMapMem.SetupLast(SL: TStringList; const AName,
  AParent, Prefix: String);
begin
   //
end;

procedure TKOLMapMem.AssignEvents(SL: TStringList; const AName: String);
begin
  inherited;
  DoAssignEvents( SL, AName,
  [ 'OnAppListChange', 'OnCloseLast', 'OnCloseMap', 'OnOpenFirst', 'OnOpenMap', 'OnUpdate' ],
  [ @OnAppListChange , @OnCloseLast , @OnCloseMap , @OnOpenFirst , @OnOpenMap , @OnUpdate  ]);
end;

procedure Register;
begin
	RegisterComponents('KOL', [TKOLMapMem]);
end;

end.

